#lang racket
(require 2htdp/universe)

;;Maximum player count
(define NUM_PLAYERS 2)

;;Boards:
(define empty_board (make-vector 20 ""))
(define start_board (make-vector 20 "?"))

(define items '("kiwi" "birne" "kirsche" "blaubeere" "johannisbeere" "kokosnuss" "orange" "wassermelone" "apfel" "graipefruit"))
(define real_board (list->vector (shuffle (append items items))))

;;Universe representation
(define UNIVERSE0 
  (list 'wait '() '(0 0) empty_board))

;;Quick accessors for the universe
(define (current_state univ)
  (first univ))

;;Connected worlds
(define (current_worlds univ)
  (second univ))
(define (world1 univ)
  (first (current_worlds univ)))
(define (world2 univ)
  (second (current_worlds univ)))

;;Counts for the worlds
(define (current_counts univ)
  (third univ))
(define (world1_count univ)
  (first (current_counts univ)))
(define (world2_count univ)
  (second (current_counts univ)))

;;The board with all the cards
(define (current_board univ)
  (fourth univ))

;;Fügt eine neue Welt hinzu 
(define (add-world univ wrld)
  (cond 
         ;;Maximale Anzahl an Spielern erreicht
         ;; --> Weise diese Welt ab
         [(= (length (current_worlds univ)) NUM_PLAYERS)
          (make-bundle univ
                       (list (make-mail wrld (list 'rejected '("dummy" 0) '("dummy" 0) (vector->list empty_board))))
                       (list wrld))]
         
         ;;Maximale Anzahl an Spielern mit dieser Welt erreicht
         ;; --> Füge die Welt zu den bekannten hinzu
         ;; --> Starte das Spiel
         [(= (length (current_worlds univ)) (- NUM_PLAYERS 1))
          (make-bundle (list
                        'play
                        (append (current_worlds univ) (list wrld))
                        '(0 0)
                        start_board)
                       (list (make-mail (world1 univ) (list 'play
                                                            (list (iworld-name (world1 univ)) 0)
                                                            (list (iworld-name wrld)          0)
                                                            (vector->list start_board)))
                             (make-mail wrld   (list 'wait
                                                     (list (iworld-name wrld)          0)
                                                     (list (iworld-name (world1 univ)) 0)
                                                     (vector->list start_board))))
                       '())]
         
         ;;Maximale Anzahl an Spielern noch nicht erreicht
         ;; --> Füge die Welt zu den bekannten hinzu
         [else 
          (make-bundle (list 
                        'wait 
                        (append (current_worlds univ) (list wrld))
                        '(0 0)
                        empty_board)
                       (list (make-mail wrld (list 'wait  (list (iworld-name wrld) 0) '("N.N." 0) (vector->list empty_board))))
                       '())]))

;;Nachrichtenaustausch zwischen den Welten 
(define (handle-messages univ wrld m)
  (cond 
    ;;Das Spiel gilt als beendet -> Anfrage eines Neustarts durch eine Welt
    [(and (equal? (current_state univ) 'finished) 
          (equal? m 'restart))
     (make-bundle (list
                   'play
                   (reverse (current_worlds univ))
                   '(0 0)
                   empty_board)
                  (list (make-mail (world1 univ) (list 'wait
                                                       (list (iworld-name (world1 univ)) 0)
                                                       (list (iworld-name (world2 univ)) 0)
                                                       (vector->list empty_board)))
                        (make-mail (world2 univ) (list 'play 
                                                       (list (iworld-name (world2 univ)) 0)
                                                       (list (iworld-name (world1 univ)) 0)
                                                       (vector->list empty_board))))
                  '())]
    ;;Eine Welt möchte ein Feld umdrehen, dazu schickt sie (select FELD_NR) an das Universum
    ;;Prüfe ob: 1. Nachricht im richtigen Format
    ;;          2. Welt gerade spielen darf
    ;;          3. Feld noch "?" ist
    ;;          4. Hat die Welt bisher max. 1 Feld umgedreht?
    [(and (list? m) (= (length m) 2) (equal? (first m) 'select)         ;; 1.
          (iworld=? wrld (world1 univ))                                 ;; 2.
          (string=? (vector-ref (current_board univ) (second m)) "?")   ;; 3.
          (< (vector-count (lambda (x) (and (not (string=? x "?"))       ;; 4.
                                           (not (string=? x ""))))
                           (current_board univ)) 2))
         ;;Decke die Karte auf -> neues Board
         (let* ([new_board (vector-copy  (current_board univ))]
                [foo (vector-set! new_board (second m) (vector-ref real_board (second m)))])
               ;;Übermittle das neue Board
               (make-bundle (list 
                             'play 
                             (current_worlds univ)
                             (current_counts univ)
                             new_board)
                            (list (make-mail (world1 univ) (list 'play
                                                                 (list (iworld-name (world1 univ)) (world1_count univ))
                                                                 (list (iworld-name (world2 univ)) (world2_count univ))
                                                                 (vector->list new_board)))
                                  (make-mail (world2 univ) (list 'wait
                                                                 (list (iworld-name (world1 univ)) (world1_count univ))
                                                                 (list (iworld-name (world2 univ)) (world2_count univ))
                                                                 (vector->list new_board))))
                            '()))]
    ;;Eine Welt möchte nach zwei Feldern weiter machen. Dazu klickt sie auf ein bereits umgedrehtes Feld erneut und
    ; schickt somit (select FELD_NR) an das Universum
    ;;Prüfe ob: 1. Nachricht im richtigen Format
    ;;          2. Welt gerade spielen darf
    ;;          3. Feld weder "?" noch "" ist
    ;;          4. Hat die Welt bisher 2 Felder umgedreht?
    [(and (list? m) (= (length m) 2) (equal? (first m) 'select)         ;; 1.
          (iworld=? wrld (world1 univ))                                 ;; 2.
          (not (string=? (vector-ref (current_board univ) (second m)) "?"))   ;; 3.
          (not (string=? (vector-ref (current_board univ) (second m)) ""))   ;; 3.
          (= (vector-count (lambda (x) (and (not (string=? x "?"))       ;; 4.
                                            (not (string=? x ""))))
                           (current_board univ)) 2))
         ;;Filtere die zwei umgedrehten Karten aus dem Spiel
         (let* ([filtered_board (vector-filter (lambda (x) (and (not (string=? x "?"))
                                                                (not (string=? x ""))))
                                               (current_board univ))])
                ;;Teste sie auf Gleichheit
                (if (and (= 2 (vector-length filtered_board))
                         (string=? (vector-ref filtered_board 0) (vector-ref filtered_board 1)))
                    ;;Ja, neuen Spielstand und neues Spielbrett berechnen
                    (let* ([new_board (vector-map (lambda (x)
                                                   (if (string=? x (vector-ref filtered_board 0))
                                                       ""
                                                       x))
                                                 (current_board univ))])
                      ;;Nun könnte jemand gewonnen haben, wir testen:
                      (if (= (vector-length (vector-filter-not (curry string=? "")  new_board)) 0)
                          ;;Es gibt nur noch umgedrehte Felder, das Spiel ist vorbei!
                          (let ([player1_count (+ (world1_count univ) 1)]
                                [player2_count (world2_count univ)])
                            (cond
                              ;;Spieler1 hat gewonnen
                              [(> player1_count player2_count)
                                (make-bundle (list 
                                              'finished 
                                              (current_worlds univ) 
                                              (list (+ (world1_count univ) 1) (world2_count univ))
                                              new_board)
                                             (list (make-mail (world1 univ) (list 'won
                                                                                  (list (iworld-name (world1 univ)) (+ (world1_count univ) 1))
                                                                                  (list (iworld-name (world2 univ)) (world2_count univ))
                                                                                  (vector->list new_board)))
                                                   (make-mail (world2 univ) (list 'lost
                                                                                  (list (iworld-name (world1 univ)) (+ (world1_count univ) 1))
                                                                                  (list (iworld-name (world2 univ)) (world2_count univ))
                                                                                  (vector->list new_board))))
                                             '())]
                              
                              ;;Spieler2 hat gewonnen
                              [(> player2_count player1_count)
                                (make-bundle (list 
                                              'finished 
                                              (current_worlds univ) 
                                              (list (+ (world1_count univ) 1) (world2_count univ))
                                              new_board)
                                             (list (make-mail (world1 univ) (list 'lost
                                                                                  (list (iworld-name (world1 univ)) (+ (world1_count univ) 1))
                                                                                  (list (iworld-name (world2 univ)) (world2_count univ))
                                                                                  (vector->list new_board)))
                                                   (make-mail (world2 univ) (list 'won
                                                                                  (list (iworld-name (world1 univ)) (+ (world1_count univ) 1))
                                                                                  (list (iworld-name (world2 univ)) (world2_count univ))
                                                                                  (vector->list new_board))))
                                             '())]
                              ;;Remis
                              [(= player2_count player1_count)
                                (make-bundle (list 
                                              'finished 
                                              (current_worlds univ) 
                                              (list (+ (world1_count univ) 1) (world2_count univ))
                                              new_board)
                                             (list (make-mail (world1 univ) (list 'remis
                                                                                  (list (iworld-name (world1 univ)) (+ (world1_count univ) 1))
                                                                                  (list (iworld-name (world2 univ)) (world2_count univ))
                                                                                  (vector->list new_board)))
                                                   (make-mail (world2 univ) (list 'remis
                                                                                  (list (iworld-name (world1 univ)) (+ (world1_count univ) 1))
                                                                                  (list (iworld-name (world2 univ)) (world2_count univ))
                                                                                  (vector->list new_board))))
                                             '())]))
                      ;;Nein, das Spiel ist noch nicht beendet!
                      ;;Übermittle das neue Board, der aktuelle Spieler bleibt dran!
                      (make-bundle (list 
                                    'play 
                                    (current_worlds univ)
                                    (list (+ (world1_count univ) 1) (world2_count univ))
                                    new_board)
                                   (list (make-mail (world1 univ) (list 'play
                                                                        (list (iworld-name (world1 univ)) (+ (world1_count univ) 1))
                                                                        (list (iworld-name (world2 univ)) (world2_count univ))
                                                                        (vector->list new_board)))
                                         (make-mail (world2 univ) (list 'wait
                                                                        (list (iworld-name (world1 univ)) (+ (world1_count univ) 1))
                                                                        (list (iworld-name (world2 univ)) (world2_count univ))
                                                                        (vector->list new_board))))
                            '())))
                    ;;Nein, die beiden Felder sind nicht gleich, neuen Spielstand und neues Spielbrett berechnen
                    (let* ([new_board (vector-map (lambda (x)
                                                   (if (string=? x "")
                                                       ""
                                                       "?"))
                                                 (current_board univ))])
                      ;;Übermittle das neue Board, der nächste Spieler ist dran!
                      (make-bundle (list 
                                    'play 
                                    (reverse (current_worlds univ))
                                    (reverse (current_counts univ))
                                    new_board)
                                   (list (make-mail (world1 univ) (list 'wait
                                                                        (list (iworld-name (world1 univ)) (world1_count univ))
                                                                        (list (iworld-name (world2 univ)) (world2_count univ))
                                                                        (vector->list new_board)))
                                         (make-mail (world2 univ) (list 'play
                                                                        (list (iworld-name (world1 univ)) (world1_count univ))
                                                                        (list (iworld-name (world2 univ)) (world2_count univ))
                                                                        (vector->list new_board))))
                            '()))))]
    ;;Sonstige Anfragen verändern das Universum nicht
    [else (make-bundle univ '() '())]))


;;Erschafft ein Universum
(universe UNIVERSE0
          (on-new add-world)
          (on-msg handle-messages))