#lang racket

;;Karten aus: https://madoo.net/39950/bildkarten-obst/

(require 2htdp/image)
(require 2htdp/universe)

;;Zustände:
;; 1. w=(status (my_name my_count) (opponent_name opponent_count)  ("?" "?" ... "?"))
;;
;; status := wait | play | won | lost | remis


;;Startzustand
(define WORLD0 (list 'wait '("Player 1" 0) '("Player 2" 0) (make-list 20 "?")))

(define (status w)
  (first w))

(define (myName w)
  (first (second w)))
(define (opponentName w)
  (first (third w)))

(define (myPairs w)
  (second (second w)))
(define (opponentPairs w)
  (second (third w)))

;;Falls ein korrekter Weltzustand empfangen wird
;;   --> setze die Welt auf die empfangene Nachricht
;;sonst: verharre im alten Weltzustand
(define (receive w m)
  (if (and (list? m) (= (length m) 4)
           (symbol? (first m)) 
           (list? (second m)) (= (length (second m)) 2)
           (list? (third m))  (= (length (third m)) 2)
           (list? (fourth m))  (= (length (fourth m)) 20))
      m
      w))


;;Zeichnen einer Welt
;;Hilfsfunktionen
(define new_game_text (above(text "Für ein neues Spiel" 16 'black)
                            (text "bitte klicken" 16 'black)))
(define (field->image str)
  (cond [(string=? str "?")
         (overlay (text str 36 'black)
                  (rectangle 100 100 'solid 'lightblue))]
        [(string=? str "")
         (rectangle 100 100 'solid 'white)]
        [else (crop 0 0 100 100 (bitmap/file (string-append str ".png")))]))

(define (fields->images xs) 
  (map field->image xs))

;;Eigentliches Zeichen
(define (draw w)
  (let ([stand (format "Aktueller Spielstand: ~a: ~a Paar~a, ~a: ~a Paar~a."
                       (myName w)
                       (myPairs w)
                       (if (or (= (myPairs w) 0) (> (myPairs w) 1)) "e" "")
                       (opponentName w)
                       (opponentPairs w)
                       (if (or (= (opponentPairs w) 0) (> (opponentPairs w) 1)) "e" ""))])
      (cond ;;Won
             [(equal? (status w) 'won)
              (above
               (text "Gewonnen" 22 'darkgreen)
               (text stand 14 'darkgreen)
               new_game_text)]
             ;;Lost
             [(equal? (status w) 'lost)
              (above 
               (text "Verloren" 22 'red)
               (text stand 14 'red)
                new_game_text)]
             ;;Remis
             [(equal? (status w) 'remis)
              (above 
               (text "Remis" 22 'blue) 
               (text stand 14 'blue) 
               new_game_text)]
             ;;Sonst wird normal gespielt
             [else
              (let ((cards (fields->images (fourth w))))
                (above
                 (beside (list-ref cards  0) (list-ref cards  1) (list-ref cards  2) (list-ref cards  3) (list-ref cards  4))
                 (beside (list-ref cards  5) (list-ref cards  6) (list-ref cards  7) (list-ref cards  8) (list-ref cards  9))
                 (beside (list-ref cards 10) (list-ref cards 11) (list-ref cards 12) (list-ref cards 13) (list-ref cards 14))
                 (beside (list-ref cards 15) (list-ref cards 16) (list-ref cards 17) (list-ref cards 18) (list-ref cards 19))
                 (rectangle 20 20 'solid 'white)
                 (if (equal? (status w) 'wait)
                     (above (text "Warte auf Gegner..." 16 'red)
                 (rectangle 10 10 'solid 'white)
                            (text stand 14 'red))
                     (above (text "Bitte Zelle markieren!" 16 'darkgreen)
                 (rectangle 10 10 'solid 'white)
                            (text stand 14 'darkgreen))))
                 )])))


;;Maus-Interaktionen
(define (handle-mouse w x_pos y_pos mouse_event)
    (if(mouse=? mouse_event "button-up")
       (if (or (equal? (status w) 'won)
               (equal? (status w) 'lost)
               (equal? (status w) 'remis))
           (make-package w 'restart)
           (let* ((column (quotient x_pos 100))
                  (row    (quotient y_pos 100))
                  (index  (+ (* 5 row) column)))
             (if (< -1 index 20) 
                 (make-package w (list 'select index))
                 w)))
       w))

;;Erstelle eine Welt und verbinde sie mit dem LOCALHOST Server
(define (create-world player_name)
    (big-bang WORLD0
             (on-receive receive)
             (to-draw draw)
             (on-mouse handle-mouse)
             (name player_name)
             (state #t)
             (register LOCALHOST)))

;;Macht zwei Welten auf
(launch-many-worlds 
  (create-world "Player 1")
  (create-world "Player 2"))